
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "Vcam.h"

#include <yarp/os/all.h>
#include <yarp/sig/all.h>
#include <yarp/dev/all.h>

#include <unistd.h>

using namespace yarp::os;
using namespace yarp::sig;
using namespace yarp::dev;

#include "drivers.h"

#include "Effects.h"

static int exitRequested = 0;

void sig_handler(int signo) {
  exitRequested = 1;
}

class VcamCommon : public Vcam {
private:
  Network yarp;
  DriverCollection dev;
  PolyDriver source;
  IFrameGrabberImage *grabber;
  ImageOf<PixelRgb> cache, proc;
  Bottle sources;
  PolyDriver destDriver;
  ConstString outputName;
  ConstString sourceName;
  int vformat;
  int w, h;
  int hout;
  int readImage;
  char *vidOut;
  ImageOf<PixelBgr> src, dest;
  ImageOf<PixelRgb> src2;
  bool output;

public:
  VcamCommon() {
    sourceName = "none";
    outputName = "none";
    open("test");
    output = false;
  }

  virtual ~VcamCommon() {
    stopOutput();
  }

  virtual bool startFileOutput(const char *name) {
    stopOutput();
    Property p;
    p.put("device", "ffmpeg_writer");
    p.put("out",name);
    p.put("framerate", 25);  // random number
    output = destDriver.open(p);
    return output;
  }

  virtual bool startDevOutput(const char *name /* = "/dev/video0" */) = 0;

  virtual bool stopOutput() = 0;

  bool open(const char *name) = 0;

  virtual bool isImage() {
    if (grabber==NULL) {
      return false;
    }
    return true; // not quite true...
  }

  virtual yarp::sig::Image *getImage() {
    if (grabber==NULL) return false;
    bool result = grabber->getImage(cache);
    if (!result) return NULL;
    if (!output) {
      Effects::apply(cache,proc);
      return &proc;
    } else {
      if (destDriver.isValid()) {
	Effects::apply(cache,proc);
	IFrameWriterImage *writer = NULL;
	destDriver.view(writer);
	writer->putImage(proc);
      } else {
	Effects::apply(cache,dest);
	if (put_image(hout, vidOut, w, h)==0) {
	  printf("*** FAILED to send image\n");
	}
      }
      return &dest;
    }
  }

  virtual bool getImage(ImageOf<PixelRgb>& img) {
    if (grabber==NULL) return false;
    bool result = grabber->getImage(cache);
    if (!result) return false;
    Effects::apply(cache,img);
    if (output) {
      if (destDriver.isValid()) {
	IFrameWriterImage *writer = NULL;
	destDriver.view(writer);
	writer->putImage(img);
      } else {
	dest.copy(img);
	if (put_image(hout, vidOut, w, h)==0) {
	  printf("*** FAILED to send image\n");
	}
      }
    }
    return true;
  }

  virtual bool close() {
    source.close();
  }

  virtual yarp::os::Bottle getSources() {
    sources.clear();
    sources.addString("test");
    for (int i=0; i<10; i++) {
      char buf[256];
      sprintf(buf,"/dev/video%d", i);
      struct stat s;
      int r = stat(buf,&s);
      if (r>=0) {
	sources.addString(buf);
	printf("adding source %s\n", buf);
      }
    }
    return sources;
  }

  virtual void setSource(const char *name) {
    if (sources.size()<1) {
      getSources();
    }
    printf("Should switch source to %s\n", name);
    open(name);
  }


  virtual yarp::os::Bottle getOutputs() {
    yarp::os::Bottle b;
    b.addString("none");
    for (int i=0; i<10; i++) {
      char buf[256];
      sprintf(buf,"/dev/video%d", i);
      struct stat s;
      int r = stat(buf,&s);
      if (r>=0) {
	b.addString(buf);
	printf("adding possible output %s\n", buf);
      }
    }
    //b.addString("capture.mpeg2video");
    return b;
  }

  virtual bool setOutput(const char *name) {
    printf("Should set output to %s\n", name);
    if (ConstString(name)=="none") {
      stopOutput();
    } else if (name[0]=='c') {
      stopOutput();
      startFileOutput(name);      
    } else {
      stopOutput();
      startDevOutput(name);
    }
    outputName = name;
    return true;
  }


  virtual yarp::os::ConstString getCurrentOutput() {
    return outputName;
  }

  virtual yarp::os::ConstString getCurrentSource() {
    return sourceName;
  }

};

